# -*- coding: utf-8 -*-
"""mercat2_kmers.py: Module for calculating k-mer counts
"""

import gzip
from pathlib import Path

## Calculate k-mer count
#
def calculateKmerCount(seq:str, kmer:int) -> dict:
    '''Calculates the k-mer count in a sequence.
    This is a helper method for find_kmers()

    Parameters:
        seq (str): The sequence to search for k-mers.
        kmer (int): The k-mer size

    Returns:
        dict: A dictionary with the counts of each k-mer found.
    '''

    kmerlist = dict()
    for i in range((len(seq) - kmer) + 1):
        k = seq[i:i+kmer]
        if k not in kmerlist:
            kmerlist[k] = 0
        kmerlist[k] += 1
    return kmerlist


## Find k-mers
def find_kmers(file:Path, kmer:int, min_count:int):
    '''Calculates the k-mer count in a fasta file.

    Parameters:
        file (Path): path to a fasta file to scan for k-mers.
        kmer (int): k-mer length.
        min_count (int): minimum count of k-mers found to be considered significant.

    Returns:
        dict: A dictionary with the counts of each k-mer found.
    '''

    num = 0
    kmerlist = dict()

    f = gzip.open(file, 'rt') if file.suffix=='.gz' else open(file, 'r')

    seq = ""
    for line in f:
        line = line.strip()
        if line.startswith(">"):
            num += 1
            if seq:
                # count k-mers
                for i in range((len(seq) - kmer) + 1):
                    k = seq[i:i+kmer]
                    if k not in kmerlist:
                        kmerlist[k] = 0
                    kmerlist[k] += 1
                seq = ""
        else:
            seq += line.replace("*","")
    # Count last sequence
    for i in range((len(seq) - kmer) + 1):
        k = seq[i:i+kmer]
        if k not in kmerlist:
            kmerlist[k] = 0
        kmerlist[k] += 1

    f.close()

    significant_kmers = dict()
    for k,v in kmerlist.items():
        if kmerlist[k] >= min_count:
            significant_kmers[k] = v

    return significant_kmers




# -*- coding: utf-8 -*-
"""mercat2_diversity.py: Module for calculating diversity metrics
"""


from pathlib import Path
import ray
from skbio.diversity import alpha as skbio_alpha
from skbio.diversity import beta_diversity
import matplotlib.pyplot as pyplot


def compute_alpha_diversity(basename, counts_tsv, out_file):

    @ray.remote(num_cpus=1)
    def get(func, count):
        try:
            method = getattr(skbio_alpha, func)
            return func, method(count)
        except:
            return func, 'NA'

    counts = list()
    with open(counts_tsv) as reader:
        reader.readline() # skip header
        for line in reader:
            counts += [int(line.split()[1])]

    jobs = list()
    for func in ['shannon', 'simpson', 'simpson_e', 'goods_coverage', 'fisher_alpha', 'dominance', 'chao1', 'chao1_ci', 'ace']:
        jobs += [get.remote(func, counts)]

    results = dict()
    while jobs:
        ready,jobs = ray.wait(jobs)
        if ready:
            key,value = ray.get(ready[0])
            results[key] = value

    with open(out_file, 'w') as writer:
        print('Metric', basename, sep='\t', file=writer)
        for func in ['shannon', 'simpson', 'simpson_e', 'goods_coverage', 'fisher_alpha', 'dominance', 'chao1', 'chao1_ci', 'ace']:
            if func not in results:
                continue
            if type(results[func]) is not str:
                try:
                    value = round(results[func], 2)
                except:
                    value = [round(x,2) for x in results[func]]
            else:
                value = results[func]
            print(func, value, sep='\t', file=writer)
    return


def compute_beta_diversity(basename, counts_tsv:Path, outpath:Path):
    outpath.mkdir(0o777, True, True)

    IDs = list()
    counts = list()
    with open(counts_tsv) as reader:
        reader.readline()
        for line in reader:
            line = line.rstrip('\n').split('\t')
            IDs.append(line[0])
            counts.append([int(x) for x in line[1:]])

    for metric in [
            "euclidean",
            "cityblock",
            "braycurtis",
            "canberra",
            "chebyshev",
            "correlation",
            "cosine",
            "dice",
            "hamming",
            "jaccard",
            "mahalanobis", # FAILED: ValueError: The number of observations (5) is too small; the covariance matrix is singular. For observations with 1024 dimensions, at least 1025 observations are required.
            "manhattan",  # aliases to "cityblock" in beta_diversity
            "matching",
            "minkowski",
            "rogerstanimoto",
            "russellrao",
            "seuclidean",
            "sokalmichener",
            "sokalsneath",
            "sqeuclidean",
            "yule",
        ]:
        try:
            distance = beta_diversity(metric, counts, IDs)

            with open(Path(outpath, f"{metric}-{basename}.tsv"), 'w') as writer:
                print('', *IDs, sep='\t', file=writer)
                for i,row in enumerate(distance.data):
                    print(IDs[i], *row, sep='\t', file=writer)
            fig = distance.plot()
            fig.savefig(Path(outpath, f"{metric}-{basename}.png"))
            pyplot.close(fig)
        except Exception as e:
            print(f'Error with beta metric: {metric.capitalize()}')
            print(e.with_traceback(None))

    return



# -*- coding: utf-8 -*-
"""mercat2_figures.py: Module for creating tables, figures
"""

import os
from pathlib import Path
import psutil
import re
import base64
import gzip
import pkg_resources as pkg
import pandas as pd
import plotly.graph_objs as go
import plotly.express as px
from sklearn.decomposition import PCA
from sklearn.decomposition import IncrementalPCA as iPCA
from dominate.tags import *
from dominate.util import raw
import timeit

from mercat2_lib import mercat2_metrics


# Global Stylesheet
STYLESHEET = pkg.resource_stream('mercat2_lib', 'data/style.css').read().decode()
LOGO = pkg.resource_stream('mercat2_lib', 'data/logo.jpg').read()
LOGO = base64.b64encode(LOGO).decode()

# TODO: Option for how to include plotly.js.
# False uses script in <head>, 'cdn' loads from internet. # Can I use both???
PLOTLY_SOURCE = 'cdn'


# RAM Usage
def mem_use():
    return round(psutil.virtual_memory().used/1024.0**3, 2)


# k-mer summary
def kmer_summary(tsv_file: dict):
    '''Creates a plotly bar graph with the k-mer counts of the top k-mers among all samples.

    Parameters: ***TODO: UPDATE THIS***
        df_all_samples (dict[name: Dataframe]): A dictionary with the Dataframes of samples containing the counts of k-mers.
        
    Returns:
        tuple: A tuple with two plotly figures: a barplot of the top k-mers and a plotly table figure labeling the k-mers.
    '''

    numKmers = 5

    top_kmers = list()
    def line_avg(str_list):
        return sum([int(x) for x in str_list[1:]]) / (len(str_list)-1)

    with open(tsv_file, 'r') as reader:
        header = reader.readline().strip().split(sep='\t')
        for line in reader:
            line = line.strip().split(sep='\t')
            if len(top_kmers) < numKmers:
                top_kmers.append(line)
            else:
                top_kmers.sort(key = lambda x: line_avg(x), reverse=False) # sort in ascending order by first element
                if line_avg(line) > line_avg(top_kmers[0]):
                    top_kmers[0] = line

    df = pd.DataFrame(top_kmers, columns=header).set_index('k-mer', drop=True).astype(int)
    df = df.T.reset_index().rename(columns=dict(index='sample'))
    df = df.melt(id_vars=['sample'], var_name='k-mer', value_name='count')#.sort_values(by='count', ascending=False)
    df['label'] = pd.Categorical(df['k-mer']).codes + 1
    df['label'] = df['label'].apply(lambda x: f'k-mer-{x}')
    df.sort_values(by=['label','count'], inplace=True, ascending=[True,False])

    fig = px.bar(df, x='sample', y='count', text='count', color='sample', facet_row='label', template="plotly_white")
    fig.for_each_annotation(lambda a: a.update(text=a.text.replace('label=', '')))
    fig.update_layout(font=dict(color="Black"))
    
    df = df[['k-mer', 'label']].drop_duplicates()

    # Add table to label kmers
    figTable = go.Figure(
        data=[go.Table(cells=dict(values=[df['label'].tolist(), df['k-mer'].tolist()]))],
        layout=go.Layout(margin=go.layout.Margin(l=0, r=0, b=0, t=0,
        ))
    )
    figTable.update_layout(height=100, font=dict(color="Black"), template='plotly_white')

    return (fig, figTable)


# GC Plot kmers
def GC_plot_kmer(df_all_samples:dict):
    '''Creates a plotly bar graph with the GC content of k-mers.

    Parameters:
        df_all_samples (dict[name: Dataframe]): A dictionary with the Dataframes of samples containing the counts of k-mers.

    Returns:
        plotly fig: A plotly barplot figure of the GC% Content of the top k-mers.
    '''

    #print("\nGC PLOT\n")
    df_list = []
    for name,value in df_all_samples.items():
        df_list.append(value['Count'].rename(name))
    df_merged = pd.concat(df_list, axis=1, keys=[s.name for s in df_list]).fillna(0)
    df_merged['Total'] = df_merged.sum(axis=1)
    df_merged['Mean'] = df_merged[[s.name for s in df_list]].mean(axis='columns')
    df_merged.sort_values(by='Mean', axis='index', ascending=False, inplace=True)
    #print(df_merged)
    df = df_merged.reset_index().rename(columns=dict(index='k-mer'))['k-mer'].iloc[0:2]
    def calc_gc(kmer:str):
        return 100.0 * (kmer.count('G') + kmer.count('C')) / len(kmer)
    df = pd.concat([df, df.apply(calc_gc)], axis=1, keys=['k-mer', 'GC'])
    #print("GC CONTENT\n", df)
    fig = px.bar(df, x='k-mer', y='GC', text='GC', template="plotly_white")
    fig.update_layout(font=dict(color="Black"))
    return fig


# GC Plot samples
def GC_plot_sample(gc_content: dict):
    '''Creates a plotly bar graph of the GC content.

    Parameters:
        gc_content (dict[str: float]): A dictionary with GC content data.

    Returns:
        plotly fig: A plotly barplot figure of the GC% Content from the gc_content dictionary.
    '''

    df = pd.DataFrame.from_dict(data=gc_content, orient='index', columns=['GC Content'])
    fig = px.bar(df, template="plotly_white",
        labels={'index':'Sample', 'value':'GC percent'})
    fig.update_layout(font=dict(color="Black"))
    return fig


# Protein Metrics Plot samples
def plot_sample_metrics(protein_samples: dict, tsv_out):
    '''Creates a plotly bar graph of the protein metrics.

    Parameters:
        protein_list (dict[str: list[str]]): A dictionary with lists of sequence files

    Returns:
        plotly fig: A plotly barplot figure of the protein metrics from the samples dictionary.
    '''

    with open(tsv_out, 'w') as writer:
        print('Sample', 'seq_name', 'length', 'PI', 'MW', 'Hydro', sep='\t', file=writer)

    figures = dict()
    for basename,files in protein_samples.items():
        for file in files:
            dfMetrics = pd.DataFrame()
            reader = gzip.open(file, 'rt') if Path(file).suffix=='.gz' else open(file, 'r')
            line = reader.readline()
            while line:
                line = line.strip()
                line = line.rstrip('*')
                if line.startswith('>'):
                    name = line[1:]
                    sequence = ""
                    line = reader.readline()
                    while line:
                        line = line.strip()
                        line = line.rstrip('*')
                        if line.startswith('>'):
                            break
                        sequence += line
                        line = reader.readline()
                    if len(sequence):
                        dfMetrics.at[name, 'Name'] = name.split()[0]
                        dfMetrics.at[name, 'Length'] = len(sequence)
                        dfMetrics.at[name, 'PI'] = mercat2_metrics.predict_isoelectric_point_ProMoST(sequence)
                        dfMetrics.at[name, 'MW'] = mercat2_metrics.calculate_MW(sequence)
                        dfMetrics.at[name, 'Hydro'] = mercat2_metrics.calculate_hydro(sequence)
                    else:
                        print("WARNING: Empty Sequence:", basename, name, sequence)
                    continue #already got next line, next item in loop
                line = reader.readline()
            reader.close()

            dfMetrics.sort_values(by='Length', ascending=False, inplace=True)
            dfMetrics.to_csv(tsv_out, sep='\t', mode='a', index=True, header=False)

            figures[f"{basename}_PI"] = px.bar(dfMetrics, x='Length', y='PI', template="plotly_white",)
            figures[f"{basename}_PI"].update_layout(font=dict(color="Black"))

            figures[f"{basename}_MW"] = px.bar(dfMetrics, x='Length', y='MW', template="plotly_white",)
            figures[f"{basename}_MW"].update_layout(font=dict(color="Black"))

            figures[f"{basename}_Hydro"] = px.bar(dfMetrics, x='Length', y='Hydro', template="plotly_white",)
            figures[f"{basename}_Hydro"].update_layout(font=dict(color="Black"))

#            dfMetrics = dfMetrics.melt(id_vars=['Name','Length'], var_name='Metric', value_name='Value')
#            figures[f"{basename}_Metrics"] = px.bar(dfMetrics, x='Length', y='Value', facet_row='Metric', template="plotly_white")
#            figures[f"{basename}_Metrics"].update_yaxes(matches=None)
#            figures[f"{basename}_Metrics"].update_layout(font=dict(color="Black"))

    return figures


# PCA
def plot_PCA(tsv_file:str, out_path:str, lowmem=None, class_file=None, DEBUG=False):
    '''Creates a 3D plotly scatter plot of the PCA of the given TSV File.

    Parameters:
        tsv_file (str): path to a TSV file with samples and kmer counts
        out_path (str): path to folder to save the PCA image and transformed PCA components

    Returns:
        plotly fig: A plotly 3D scatter plot of the PCA of the dataframe.
    '''

    start_time = timeit.default_timer()

    pca_tsv = os.path.join(out_path, 'pca.tsv')
    chunk_size = 1000
    names = list()
    with open(tsv_file) as reader:
        reader.readline() # skip header
        for line in reader:
            names.append(re.sub(r'_protein', '', line.split()[0]))
    if DEBUG:
        print(f"\nTime to get name list: {round(timeit.default_timer() - start_time,2)} seconds")
        print(f"Virtual Memory {mem_use()}GB")

    if lowmem is None and len(names) > chunk_size:
        lowmem = True
    else:
        lowmem = False

    print("Using Incremental PCA:", lowmem)
    
    name_iter = iter(names)
    pca = iPCA(n_components=3, batch_size=100) if lowmem else PCA(n_components=3)
    XDF = pd.read_csv(tsv_file, sep='\t', index_col=0, chunksize=chunk_size) if lowmem else pd.read_csv(tsv_file, sep='\t', index_col=0)
    if DEBUG:
        print(f"\nTime to read TSV file: {round(timeit.default_timer() - start_time,2)} seconds")
        print(f"Virtual Memory {mem_use()}GB")
    
    if lowmem:
        # Incremental PCA
        # Partial Fit
        for chunk in XDF:
            pca.partial_fit(chunk)
        if DEBUG:
            print(f"\nTime for iPCA partial_fit: {round(timeit.default_timer() - start_time,2)} seconds")
            print(f"Virtual Memory {mem_use()}GB")

        # Transform
        XDF = pd.read_csv(tsv_file, sep='\t', index_col=0, chunksize=chunk_size)
        with open(pca_tsv, 'w') as pca_out:
            print('sample', 'PC1', 'PC2', 'PC3', sep='\t', file = pca_out)
            for chunk in XDF:
                for row in pca.transform(chunk):
                    pca_out.write(next(name_iter))
                    for c in row:
                        pca_out.write(f'\t{c}')
                    pca_out.write('\n')
        if DEBUG:
            print(f"\nTime for iPCA transform: {round(timeit.default_timer() - start_time,2)} seconds")
            print(f"Virtual Memory {mem_use()}GB")
    else:
        # Standard PCA
        XDF = pca.fit_transform(XDF)
        if DEBUG:
            print(f"\nTime to compute PCA fit_transform: {round(timeit.default_timer() - start_time,2)} seconds")
            print(f"Virtual Memory {mem_use()}GB")

        with open(pca_tsv, 'w') as pca_out:
            print('sample', 'PC1', 'PC2', 'PC3', sep='\t', file = pca_out)
            for row in XDF:
                pca_out.write(next(name_iter))
                for c in row:
                    pca_out.write(f'\t{c}')
                pca_out.write('\n')
        if DEBUG:
            print(f"Time to save PCA TSV file: {round(timeit.default_timer() - start_time,2)} seconds")
            print(f"Virtual Memory {mem_use()}GB")


    start_time = timeit.default_timer()
    XDF = pd.read_csv(pca_tsv, sep='\t', index_col=0)

    if class_file:
        df_tax = pd.read_csv(class_file, sep='\t', index_col=0, names=['class'])
        XDF['class'] = XDF.index.map(df_tax['class']).fillna('NA')
        color_col = 'class'
        if DEBUG:
            print(f"\nTime to load CLASS file: {round(timeit.default_timer() - start_time,2)} seconds")
            print(f"Virtual Memory {mem_use()}GB")
    else:
        color_col = names

    # Plotly PCA
    start_time = timeit.default_timer()
    labels_axis = {
        f'PC{i}' : f"PC {i} ({val:.1f}%)"
        for i,val in enumerate(pca.explained_variance_ratio_ * 100, start=1)
        }

    figPCA = px.scatter_3d(
        XDF, x='PC1', y='PC2', z='PC3', color=color_col,
        labels=labels_axis, template="plotly_white"
    )
    figPCA.update_layout(font=dict(color="Black"),
        margin=dict(l=0, r=0, t=0, b=0),)
    figPCA.update_layout(scene = dict(
        xaxis = dict(
            backgroundcolor="White",
            gridcolor="LightGray",
            showbackground=False,
            zerolinecolor="LightGray"
            ),
        yaxis = dict(
            backgroundcolor="White",
            gridcolor="LightGray",
            showbackground=False,
            zerolinecolor="LightGray"
            ),
        zaxis = dict(
            backgroundcolor="White",
            gridcolor="LightGray",
            showbackground=False,
            zerolinecolor="LightGray"
            ),
        ))

    figPCA.write_image(f"{out_path}/pca{'_incremental' if lowmem else ''}.png")
    
    print(f"Time to compute 3D PCA: {round(timeit.default_timer() - start_time,2)} seconds")
    if DEBUG:
        print(f"Virtual Memory {mem_use()}GB")

    start_time = timeit.default_timer()
    figPCA2d = None
    if pca.explained_variance_ratio_[2] * 100 < 1:
        figPCA2d = px.scatter(
            XDF, x='PC1', y='PC2', color=color_col,
            labels=labels_axis, template="plotly_white"
        )
        figPCA2d.update_layout(font=dict(color="Black"),
            margin=dict(l=0, r=0, t=0, b=0),)
        figPCA2d.write_image(f"{out_path}/pca2D{'_incremental' if lowmem else ''}.png")
        print(f"Time to compute 2D PCA: {round(timeit.default_timer() - start_time,2)} seconds")
        if DEBUG:
            print(f"Virtual Memory {mem_use()}GB")

    return figPCA, figPCA2d



# -*- coding: utf-8 -*-
"""mercat2_report.py: Module for creating html reports
"""

import os
import resource
import psutil
import re
import base64
import pkg_resources as pkg
import time
import dominate
from dominate.tags import *
from dominate.util import raw

from mercat2_lib import mercat2_metrics


# Global Stylesheet
STYLESHEET = pkg.resource_stream('mercat2_lib', 'data/style.css').read().decode()
LOGO = pkg.resource_stream('mercat2_lib', 'data/logo.jpg').read()
LOGO = base64.b64encode(LOGO).decode()

# TODO: Option for how to include plotly.js.
# False uses script in <head>, 'cdn' loads from internet. # Can I use both???
PLOTLY_SOURCE = 'cdn'


# RAM Usage
def mem_use():
    return round(psutil.virtual_memory().used/1024.0**3, 2)


# Write HTML Report
def write_html(outfile:str, figPlots:dict, tsv_stats:dict):
    '''Creates an html report with the figures and tsv files.

    Parameters:
        outfile (str): A path to the html file to save the report.
        figPlots (dict): A dictionary of plotly figures.
        tsv_stats (dict): A dictionary of paths to TSV files to add to the downloads section.

    Returns:
        None
    '''

    # WRITE TO FORMATED HTML
    with dominate.document(title='K-Mer Report') as doc:
        # Header
        with doc.head:
            meta(charset="utf-8")
            script(type="text/javascript", src="plotly-2.0.0.min.js")
            with style(type="text/css"):
                raw('\n'+STYLESHEET)
        # Body of HTML
        with div(cls="document", id="mercat-summary"):
            # Title bar at the top
            with h1(cls="title"):
                img(src=f"data:image/png;base64,{LOGO}", height="40")
                a("MERCAT2", cls="reference external", href="https://github.com/raw-lab/mercat2")
                raw(" - Summary")
            # Table of contents - left panel
            with div(cls="contents topic", id="contents"):
                with ul(cls="simple"):
                    li(a("Summary", cls="reference internal", href="#summary"))
                    with ul():
                        for key in figPlots.keys():
                            li(a(f"{key}", cls="reference internal", href=f"#{key}"))
                    li(a("Downloads", cls="reference internal", href="#downloads"))
            # Main panel
            with div(h1("Summary"), cls="section", id="summary"):
                for key,figures in figPlots.items():
                    with div(h2(f"{key}"), cls="section", id=f"{key}"):
                        try:
                            raw(figures.to_html(full_html=False, include_plotlyjs=PLOTLY_SOURCE))
                        except:
                            for fig in figures:
                                raw(fig.to_html(full_html=False, include_plotlyjs=PLOTLY_SOURCE))
            # Downloads at the bottom
            with div(cls="section", id="downloads"):
                h1("Downloads")
                with div(cls="docutils container", id="attachments"):
                    with blockquote():
                        with div(cls="docutils container", id="table-1"):
                            with dl(cls="docutils"):
                                for key,value in tsv_stats.items():
                                    data_URI = f"data:text/tab-separated-values;base64,{value}"
                                    dt(key)
                                    dd(a(key+".tsv", href=data_URI, download=key+".tsv", draggable="true"))
                div(time.strftime("%Y-%m-%d", time.localtime()), cls="docutils container", id="metadata")

    with open(outfile, 'w') as writer:
        writer.write(doc.render())
    return


# Merge TSV Files
def merge_tsv(tsv_list:dict, out_file:os.PathLike):
    FLIMIT = len(tsv_list) >= resource.getrlimit(resource.RLIMIT_NOFILE)[0]
    names = sorted(list(tsv_list.keys()))
    file_list = dict()
    header = ""
    for name in names:
        if FLIMIT:
            with open(tsv_list[name]) as f:
                head = f.readline() # skip header
                if not header:
                    header = head.split('\t')[0]
                file_list[name] = f.tell()
        else:
            file_list[name] = open(tsv_list[name])
            head = file_list[name].readline() # skip header
            if not header:
                header = head.split('\t')[0]
    with open(out_file, 'w') as writer:
        print(header, '\t'.join(names), sep='\t', file=writer)
        lines = dict()
        kmers = set()
        for name in names:
            if FLIMIT:
                with open(tsv_list[name]) as f:
                    f.seek(file_list[name])
                    lines[name] = f.readline().split()
                    file_list[name] = f.tell()
            else:
                lines[name] = file_list[name].readline().split()
            kmers.add(lines[name][0])
        kmer = sorted(kmers)[0]
        while True:
            line = [kmer]
            kmers = set()
            for name in names:
                if not lines[name]:
                    line.append('0')
                elif lines[name][0] > kmer:
                    line.append('0')
                else:
                    line.append(lines[name][1])
                    if FLIMIT:
                        with open(tsv_list[name]) as f:
                            f.seek(file_list[name])
                            lines[name] = f.readline().split()
                            file_list[name] = f.tell()
                    else:
                        lines[name] = file_list[name].readline().strip('\n').split('\t')
                        lines[name] = [x for x in lines[name] if len(x) > 0]
                    if lines[name]:
                        kmers.add(lines[name][0])
            print('\t'.join(line), file=writer)
            if not kmers:
                break
            kmer = sorted(kmers)[0]
    if FLIMIT:
        for name in names:
            file_list[name].close()
    return


# Merge TSV Files, Transposed
def merge_tsv_T(tsv_list:dict, out_file:os.PathLike):
    names = sorted(list(tsv_list.keys()))#, key=lambda x: x.lower())
    header = set()
    #file_list = dict()
    for name in names:
        with open(tsv_list[name]) as reader:
            #file_list[name] = open(tsv_list[name], 'r')
            #file_list[name].readline() # skip header
            reader.readline() # skip header
            for line in reader:
                kmer = line.split()[0]
                header.add(kmer)
    
    header = list(header)
    with open(out_file, 'w') as writer:
        print('sample', '\t'.join(header), sep='\t', file=writer)
        for name in names:
            with open(tsv_list[name]) as reader:
                #file_list[name].seek(0)
                #file_list[name].readline() # skip header
                reader.readline() # skip header
                counts = dict()
                #for line in file_list[name]:
                for line in reader:
                    kmer,count = line.split()
                    counts[kmer] = count
                #file_list[name].close()
            writer.write(name)
            for kmer in header:
                if kmer in counts:
                    writer.write(f'\t{counts[kmer]}')
                else:
                    writer.write(f'\t0')
            writer.write('\n')
    return



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""mercat2_Chunker.py: Class for splitting fasta and fastq files into smaller chunks
Takes sequence delimiters into account (i.e. '>', '@')
"""

import os
import gzip
from pathlib import Path
import argparse
import glob


class Chunker:

    def __init__(self, path, dest, chunksize='1000M', delim=None, lines=None):
        self.path = str(path)
        self.dest = dest
        self.chunksize = human2bytes(chunksize)
        self.delim = delim
        self.lines = lines

        self.fn = os.path.basename(self.path)
        #self.name, self.ext = os.path.splitext(self.fn)
        self.name = Path(path).stem.split('.')[0]
        self.ext = ''.join(Path(path).suffixes[:-1])

        os.makedirs(dest, exist_ok=True)

        self.stream()
        self.files = glob.glob(os.path.join(dest, '*'))

    def stream(self):
        if self.delim is not None:
            self.stream_delim()
        else:
            self.stream_lines()

    def stream_delim(self):
        i = 0
        fout = open(os.path.join(self.dest, '%s.%05d%s' % (self.name, i, self.ext)), 'w')
        inf = gzip.open(self.path, 'rt') if self.path.endswith('.gz') else open(self.path, 'r')
        #with open(self.path, 'r') as inf:
        for l in inf:
            if self.delim in l:
                # check filesize, potentially increase i
                fout.flush()
                size = os.fstat(fout.fileno()).st_size
                if size >= self.chunksize:
                    fout.close()
                    i += 1
                    fout = open(os.path.join(self.dest, '%s.%05d%s' % (self.name, i, self.ext)), 'w')
                    fout.write(l)
                else:
                    fout.write(l)
            else:
                fout.write(l)
        inf.close()
        fout.close()

    def stream_lines(self):
        i = 0
        fout = open(os.path.join(self.dest, '%s.%05d%s' % (self.name, i, self.ext)), 'w')
        with open(self.path, 'r') as inf:
            for j, l in enumerate(inf):
                if j % self.lines == 0:
                    # check filesize, potentially increase i
                    fout.flush()
                    size = os.fstat(fout.fileno()).st_size
                    if size >= self.chunksize:
                        fout.close()
                        i += 1
                        fout = open(os.path.join(self.dest, '%s.%05d%s' % (self.name, i, self.ext)), 'w')
                        fout.write(l)
                    else:
                        fout.write(l)
                else:
                    fout.write(l)
            fout.close()


def human2bytes(s):
    """
    Attempts to guess the string format based on default symbols
    set and return the corresponding bytes as an integer.
    When unable to recognize the format ValueError is raised.

      >>> human2bytes('0 B')
      0
      >>> human2bytes('1 K')
      1024
      >>> human2bytes('1 M')
      1048576
      >>> human2bytes('1 Gi')
      1073741824
      >>> human2bytes('1 tera')
      1099511627776

      >>> human2bytes('0.5kilo')
      512
      >>> human2bytes('0.1  byte')
      0
      >>> human2bytes('1 k')  # k is an alias for K
      1024
      >>> human2bytes('12 foo')
      Traceback (most recent call last):
          ...
      ValueError: can't interpret '12 foo'
    """
    SYMBOLS = {
        'customary': ('B', 'K', 'M', 'G', 'T', 'P', 'E', 'Z', 'Y'),
        'customary_ext': ('byte', 'kilo', 'mega', 'giga', 'tera', 'peta', 'exa',
                          'zetta', 'iotta'),
        'iec': ('Bi', 'Ki', 'Mi', 'Gi', 'Ti', 'Pi', 'Ei', 'Zi', 'Yi'),
        'iec_ext': ('byte', 'kibi', 'mebi', 'gibi', 'tebi', 'pebi', 'exbi',
                    'zebi', 'yobi'),
    }

    init = s
    num = ""
    while s and s[0:1].isdigit() or s[0:1] == '.':
        num += s[0]
        s = s[1:]
    num = float(num)
    letter = s.strip()
    for name, sset in list(SYMBOLS.items()):
        if letter in sset:
            break
    else:
        if letter == 'k':
            # treat 'k' as an alias for 'K' as per: http://goo.gl/kTQMs
            sset = SYMBOLS['customary']
            letter = letter.upper()
        else:
            raise ValueError("can't interpret %r" % init)
    prefix = {sset[0]: 1}
    for i, s in enumerate(sset[1:]):
        prefix[s] = 1 << (i + 1) * 10
    return int(num * prefix[letter])


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Split input file into pieces.')

    # required
    parser.add_argument('infile', help='Path to input file.')
    parser.add_argument('outfolder', help='Path to output folder.')

    # flags
    parser.add_argument('-c', '--chunksize', default='1000M', help='Approximate size of file chunks.')

    # required flag
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument('-d', '--delimiter', help='Delimiter for preserving text groups.')
    group.add_argument('-l', '--lines', type=int, help='Number of lines to be considered a text group.')

    args = parser.parse_args()

    c = Chunker(args.infile, args.outfolder, chunksize=args.chunksize, delim=args.delimiter, lines=args.lines)



import os
import re
import setuptools

main_script = 'bin/mercat2.py'
version = re.search(r'__version__\s+= "(.+)"', open(main_script).read()).group(1)
author = re.search(r'__author__\s+= "(.+)"', open(main_script).read()).group(1)


# recursively load package files
def package_files(directory):
    paths = []
    for (path, _, filenames) in os.walk(directory):
        for filename in filenames:
            if not filename.endswith('.py'):
                paths.append(os.path.join('..', path, filename))
    return paths


setuptools.setup(
    name = "mercat2",
    version = version,
    author = author,
    author_email = "jlfiguer@charlotte.edu",
    description = "versatile k-mer counter and diversity estimator for database independent property analysis (DIPA) for multi-omic analysis",
    long_description = open("README.md", "r").read(),
    long_description_content_type = "text/markdown",
    url = "https://github.com/raw-lab/mercat2",
    scripts = [main_script], # scripts to install to 'bin' path
    packages = ['mercat2_lib'], # list of packages, installed to 'site-packages' folder
    package_dir = dict(mercat2_lib='lib'), # dict with 'package'='relative dir'
    package_data = dict(mercat2_lib=package_files('lib/')), # add non-python data to package, relative paths
    license = "BSD License", # metadata
    platforms = ['Unix'], # metadata
    classifiers = [ # This is the new updated way for metadata, but old way seems to still be used in some of the output
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "License :: OSI Approved :: BSD License",
        "Operating System :: Unix",
    ],
    python_requires = '>=3.9',
    install_requires = [
        'setuptools',
        'grpcio ==1.43',
        'ray',
        'configargparse',
        'pandas',
        'numpy',
        'humanize',
        'plotly',
        'psutil',
        'dominate',
        'scikit-learn',
        'scikit-bio',
        'scipy',
        'metaomestats',
        'kaleido',

# ray-core ray-dashboard ray-default ray-tune
    ]
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""versatile k-mer counter and diversity estimator for database independent property analysis (DIPA) for multi-omic analysis"""

__version__     = "1.4.0"
__author__      = "Jose L. Figueroa III, Richard A. White III"
__copyright__   = "Copyright (c) 2022-2024"

import os
import re
from pathlib import Path
import subprocess
import shutil
import psutil
from distutils.util import strtobool
import ray
import argparse
import timeit

# Mercat libraries
from mercat2_lib import (mercat2_fasta, mercat2_Chunker, mercat2_kmers, mercat2_diversity, mercat2_figures, mercat2_report)


# GLOBAL VARIABLES
FILE_EXT_FASTQ = [".fq", ".fastq", ".fq.gz", ".fastq.gz"]
FILE_EXT_NUCLEOTIDE = [".fasta", ".fa", ".fna", ".ffn", ".fasta.gz", ".fa.gz", ".fna.gz", ".ffn.gz"]
FILE_EXT_PROTEIN = [".faa", ".faa.gz"]

# RAM Usage
def mem_use():
    return round(psutil.virtual_memory().used/1024.0**3, 2)


## Parse Arguments
#
def parseargs():
    '''Returns the parsed command line options.'''
    num_cores = psutil.cpu_count(logical=False)
    parser = argparse.ArgumentParser(formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.add_argument('-i', required=False, default=list(), help='path to input file', nargs='+')
    parser.add_argument('-f', type=str, required=False, help='path to folder containing input files')
    parser.add_argument('-k', type=int, required = True, help='kmer length')
    parser.add_argument('-n', type=int, default=num_cores, help='no of cores [auto detect]')
    parser.add_argument('-c', type=int, default=10, help='minimum kmer count [10]')
    parser.add_argument('-prod', action='store_true', help='run Prodigal on fasta files')
    parser.add_argument('-fgs', action='store_true', help='run FragGeneScanRS on fasta files')
    parser.add_argument('-s', type=int, default=100, required=False, help='Split into x MB files. [100]')
    parser.add_argument('-o', type=str, default='mercat_results', required=False, help="Output folder, default = 'mercat_results' in current directory")
    parser.add_argument('-replace', action='store_true', help="Replace existing output directory [False]")
    parser.add_argument('-lowmem', type=strtobool, default=None, help="Flag to use incremental PCA when low memory is available. [auto]")
    parser.add_argument('-skipclean', action='store_true', help='skip trimming of fastq files')
    parser.add_argument('-toupper', action='store_true', help='convert all input sequences to uppercase')
    parser.add_argument('-category_file', type=str, default=None, help=argparse.SUPPRESS)
    parser.add_argument('-debug', action='store_true', help=argparse.SUPPRESS)
    parser.add_argument('-pca', action='store_true', help='create interactive PCA plot of the samples (minimum of 4 fasta files required)')
    parser.add_argument('--version', '-v', action='version',
                        version=f'MerCat2: \n version: {__version__}',
                        help='show the version number and exit')

    # Process arguments
    args = parser.parse_args()

    global DEBUG
    DEBUG = args.debug

    # Check folder/input file options
    if not args.i and not args.f:
        parser.error("Please provide either an input file (-i) or an input folder (-f)")
    for filename in args.i:
        if not os.path.isfile(filename):
            parser.error(f"file '{filename}' is not valid.\n")
    if args.f and not os.path.isdir(args.f):
        parser.error(f"folder {args.f} is not valid.\n")

    # check prodigal/protein flags
    if args.prod:
        if not mercat2_fasta.check_command('prodigal'):
            exit(1)

    return args, parser


## Chunk Files
#
@ray.remote(num_cpus=1)
def chunk_files(name:str, filename:str, chunk_size:int, outpath:str):
    '''Checks if the given file is larger than chunk_size and splits the file as necessary.
    Fasta files are split at sequence headers to keep individual sequences contiguous.

    Parameters:
        name (str): The name of the sample file for proper grouping of the files.
        file (str): Path to a fasta.
        chunk_size (int): The size, in MB, to split the files at.
        outpath (str): Path to a folder to save the chunked files in. Folder should not exist.

    Returns:
        tuple: A tuple with the name, and list of paths to the chunked files.
    '''

    if os.stat(filename).st_size >= (chunk_size*1024*1024):
        os.makedirs(outpath, exist_ok=True)
        all_chunks = mercat2_Chunker.Chunker(filename, outpath, str(chunk_size)+"M", ">").files
    else:
        all_chunks = [filename]
    return (name, all_chunks)

@ray.remote(num_cpus=1)
def diversity(key, infile, outfile, sample_type):
    mercat2_diversity.compute_alpha_diversity(key, infile, outfile)
    return (sample_type, outfile)
@ray.remote(num_cpus=1)
def countKmers(file, kmer, min_count):
    return mercat2_kmers.find_kmers(Path(file), kmer, min_count)
@ray.remote(num_cpus=1)
def run_mercat2(basename:str, files:list, out_file:os.PathLike, kmer, min_count, num_cores):
    kmers = dict()
    jobs = []
    for file in files:
        jobs += [countKmers.remote(file, kmer, min_count)]
    while(jobs):
        ready,jobs = ray.wait(jobs)
        for k,v in ray.get(ready[0]).items():
            if k in kmers:
                kmers[k] += v
            else:
                kmers[k] = v
    if len(kmers):
        print(f"Significant k-mers: {len(kmers)}")
        with open(out_file, 'w') as writer:
            print('k-mer', f'{basename}_Count', sep='\t', file=writer)
            for kmer,count in sorted(kmers.items()):
                print(kmer, count, sep='\t', file=writer)
        return basename,out_file
    else:
        print(f"No significant k-mers found")
        return basename,None

## Create Figures
#
def createFigures(tsv_list:dict, type_string:str, out_path:os.PathLike, lowmem=None, class_file=None, pca=False):
    print(f"\nCreating {type_string} Graphs")

    figPlots = dict()
    start_time = timeit.default_timer()

    # Stacked Bar Plots (top kmer counts)
    combined_tsv = os.path.join(out_path, f'combined_{type_string}.tsv')
    if not os.path.exists(combined_tsv):
        mercat2_report.merge_tsv(tsv_list, combined_tsv)
        if DEBUG:
            print(f"\nTime to merge TSV files: {round(timeit.default_timer() - start_time,2)} seconds")
            print(f"Virtual Memory {mem_use()}GB")

    combined_T_tsv = os.path.join(out_path, f'combined_{type_string}_T.tsv')
    if not os.path.exists(combined_T_tsv):
        mercat2_report.merge_tsv_T(tsv_list, combined_T_tsv)
        if DEBUG:
            print(f"\nTime to merge TSV transposed files: {round(timeit.default_timer() - start_time,2)} seconds")
            print(f"Virtual Memory {mem_use()}GB")


    figPlots[f"Combined {type_string} kmer Summary"] = mercat2_figures.kmer_summary(combined_tsv)
    if DEBUG:
        print(f"Time to compute Combined {type_string} kmer Summary: {round(timeit.default_timer() - start_time,2)} seconds")
        print(f"Virtual Memory {mem_use()}GB")

    # PCA
    if pca and len(tsv_list) > 3:
        print("\nRunning PCA")
        start_time = timeit.default_timer()

        out_pca = os.path.join(out_path, f'pca_{type_string}')
        os.makedirs(out_pca, exist_ok=True)
        pca3d,pca2d = mercat2_figures.plot_PCA(combined_T_tsv, out_pca, lowmem, class_file, DEBUG)
        if pca3d:
            figPlots[f"{type_string} PCA 3D"] = pca3d
        if pca2d:
            figPlots[f"{type_string} PCA 2D"] = pca2d

    return figPlots


## Main
#
def mercat_main():
    '''Main method'''

    # Process Arguments
    args, parser = parseargs()
    # Set Flags
    m_kmer = args.k
    m_num_cores = args.n
    m_inputfile = args.i
    m_inputfolder = args.f
    m_min_count = args.c
    m_flag_prodigal = args.prod
    m_flag_fgs = args.fgs
    m_chunk_size = args.s
    m_outputfolder = Path(args.o)
    m_lowmem = None if args.lowmem is None else bool(args.lowmem)
    m_skipclean = args.skipclean
    m_toupper = args.toupper
    m_class_file = args.category_file
    m_pca = args.pca
    
    if m_outputfolder.exists():
        if args.replace:
            shutil.rmtree(m_outputfolder)
        else:
            parser.error(f"Output folder exists, please specify another folder or use the flag '-replace' to override the files. '{m_outputfolder}'")
    
    m_outputfolder.mkdir(0o777, True, True)

    print(f"\nStarting MerCat2 v{__version__} with k-mer {m_kmer} and {m_num_cores} threads\n")

    ray.init(num_cpus=m_num_cores, log_to_driver=DEBUG)
    if DEBUG:
        print(f"\nVirtual Memory {mem_use()}GB")

    # Load input files
    gc_content = dict()
    samples = dict(nucleotide=dict(),
                protein=dict(),
                prodigal=dict(),
                fgs=dict())

    cleanpath = os.path.join(m_outputfolder, 'clean')
    print("Loading files")
    start_time = timeit.default_timer()
    jobsQC = list()
    jobsFastq = list()
    jobsContig = list()

    @ray.remote(num_cpus=1)
    def fastq_qc(file, cleanpath, basename):
        return (basename, mercat2_fasta.qc(file, cleanpath, basename))
    @ray.remote(num_cpus=1)
    def clean_contig(file, cleanpath, basename, toupper, skipclean):
        if skipclean:
            stat = None
        else:
            file,stat = mercat2_fasta.removeN(file, cleanpath, toupper)
        return (basename, file, stat)
    @ray.remote(num_cpus=1)
    def fastq_to_fasta(file, cleanpath, basename, skiptrim:bool):
        jobsQC = [fastq_qc.remote(file, cleanpath, basename)]
        if not skiptrim:
            file = mercat2_fasta.trim(file, cleanpath, basename)
            jobsQC += [fastq_qc.remote(file, cleanpath, basename)]
        return (basename, mercat2_fasta.fq2fa(file, cleanpath, basename), jobsQC)
    # Load files from folder
    if m_inputfolder:
        m_inputfolder = os.path.abspath(os.path.expanduser(m_inputfolder))
        for fname in os.listdir(m_inputfolder):
            file = Path(m_inputfolder, fname)
            if file.is_file:
                suffixes = file.suffixes
                for i in reversed(range(len(suffixes))):
                    if ''.join(suffixes[i:]) in FILE_EXT_FASTQ + FILE_EXT_NUCLEOTIDE + FILE_EXT_PROTEIN:
                        m_inputfile.append(file)
                        break
    # Read input files
    if m_inputfile:
        for filename in m_inputfile:
            basepath = Path(filename).expanduser().absolute()
            suffixes = basepath.suffixes
            f_ext = ''
            for i in reversed(range(len(suffixes))):
                if ''.join(suffixes[i:]) in FILE_EXT_FASTQ + FILE_EXT_NUCLEOTIDE + FILE_EXT_PROTEIN:
                    f_ext = ''.join(suffixes[i:])
            basename = basepath.name.removesuffix(f_ext)
            print(basename, f_ext)
            if f_ext in FILE_EXT_FASTQ:
                jobsFastq += [fastq_to_fasta.remote(filename, cleanpath, basename, m_skipclean)]
            elif f_ext in FILE_EXT_NUCLEOTIDE:
                jobsContig += [clean_contig.remote(filename, cleanpath, basename, m_toupper, m_skipclean)]
                command = [ 'countAssembly.py', '-f', filename, '-i', '100' ]
                statfile = os.path.join(m_outputfolder, 'stats', f'{basename}.txt')
                os.makedirs(os.path.join(m_outputfolder, 'stats'), exist_ok=True)
                with open(statfile, 'w') as writer:
                    subprocess.run(command, stdout=writer, stderr=subprocess.DEVNULL)
            elif f_ext in FILE_EXT_PROTEIN:
                samples['protein'][basename] = [filename]

    # Wait for jobs
    while jobsFastq:
        ready,jobsFastq = ray.wait(jobsFastq)
        basename,file,jobsqc = ray.get(ready[0])
        samples['nucleotide'][basename] = [file]
        jobsQC += jobsqc
        #gc_content[basename] = stat['GC Content']
    while jobsContig:
        ready,jobsContig = ray.wait(jobsContig)
        basename,file,stat = ray.get(ready[0])
        samples['nucleotide'][basename] = [file]
        if stat:
            gc_content[basename] = stat['GC Content']

    print(f"Time to load {len(samples['nucleotide'])+len(samples['protein'])} files: {round(timeit.default_timer() - start_time,2)} seconds")
    if DEBUG:
        print(f"Virtual Memory {mem_use()}GB")
    start_time = timeit.default_timer()

    # Begin processing files
    figPlots = dict()
    tsv_stats = dict()
    jobsDiversity = list()
    report_dir = os.path.join(m_outputfolder, "report")
    os.makedirs(report_dir, exist_ok=True)

    ## Process Nucleotides ##
    if samples['nucleotide']:
        # Chunk Nucleotides
        if m_chunk_size > 0:
            print("Checking for large nucleotide files")
            start_time = timeit.default_timer()
            dir_chunks = os.path.join(m_outputfolder, "chunks_nucleotide")
            jobs = []
            for basename,file in samples['nucleotide'].items():
                chunk_path = os.path.join(dir_chunks, basename)
                jobs += [chunk_files.remote(basename, *file, m_chunk_size, chunk_path)]
            while jobs:
                ready,jobs = ray.wait(jobs)
                name,chunks = ray.get(ready[0])
                chunks = [Path(c) for c in chunks]
                samples['nucleotide'][name] += chunks
            if DEBUG:
                print(f"Time to check for large files: {round(timeit.default_timer() - start_time,2)} seconds")

        print("Processing Nucleotides")
        print(f"Running Mercat2 using {m_num_cores} cores")
        out_tsv = os.path.join(m_outputfolder, 'tsv_nucleotide')
        os.makedirs(out_tsv, exist_ok=True)
        jobs = []
        start_time = timeit.default_timer()
        for basename,files in samples['nucleotide'].items():
            out_counts = os.path.join(out_tsv, f"{basename}_counts.tsv")
            files = files[1:] if len(files)>1 else [files[0]]
            jobs += [run_mercat2.remote(basename, files, out_counts, m_kmer, m_min_count, m_num_cores)]
        tsv_list = dict()
        while(jobs):
            ready,jobs = ray.wait(jobs)
            basename,kmers = ray.get(ready[0])
            if kmers:
                tsv_list[basename] = kmers
        print(f"Time to count {m_kmer}-mers: {round(timeit.default_timer() - start_time,2)} seconds")
        if DEBUG:
            print(f"Virtual Memory {mem_use()}GB")
        # Stacked Bar Plots (top kmer counts)
        start_time = timeit.default_timer()
        if len(tsv_list):
            figPlots.update(createFigures(tsv_list, "Nucleotide", m_outputfolder, m_lowmem, m_class_file, m_pca))
            tsvfile = Path(m_outputfolder, f'combined_{sample_type}_T.tsv')
            sample_type = "Nucleotide"
            mercat2_diversity.compute_beta_diversity(sample_type, tsvfile, Path(m_outputfolder, "report", "diversity"))
            if DEBUG:
                print(f"Time to calculate Beta Diversity: {round(timeit.default_timer() - start_time,2)} seconds")
        for basename,filename in tsv_list.items():
            os.makedirs(os.path.join(report_dir, 'diversity'), exist_ok=True)
            outfile = os.path.join(report_dir, 'diversity', f'nucleotide-{basename}.tsv')
            jobsDiversity += [diversity.remote(basename, filename, outfile, 'Nucleotide')]
        if len(gc_content) > 0:
            figPlots['Sample GC Summary'] = mercat2_figures.GC_plot_sample(gc_content)


    ## Process Proteins ##
    # PROD ORF Call
    if m_flag_prodigal and samples['nucleotide']:
        @ray.remote(num_cpus=1)
        def orf_call_prod(basename, file, prodpath):
            return mercat2_fasta.orf_call(basename, file, prodpath)

        print(f"\nRunning Prodigal on {len(samples['nucleotide'])} files")
        start_time = timeit.default_timer()
        prodpath = os.path.join(m_outputfolder, 'prodigal')
        os.makedirs(prodpath, exist_ok=True)
        jobsProd = []
        for basename,files in samples['nucleotide'].items():
            jobsProd += [orf_call_prod.remote(basename, files[0], prodpath)]
        while jobsProd:
            ready,jobsProd = ray.wait(jobsProd)
            ret = ray.get(ready[0])
            if ret:
                samples['prod'][ret[0]] = [ret[1]]
        if DEBUG:
            print(f"Time to run Prodigal: {round(timeit.default_timer() - start_time,2)} seconds")
            print(f"Virtual Memory {mem_use()}GB")

    # FGS ORF Call
    if m_flag_fgs and samples['nucleotide']:
        @ray.remote(num_cpus=1)
        def orf_call_fgs(basename, file, outpath):
            return mercat2_fasta.orf_call_fgs(basename, file, outpath)
        print(f"\nRunning FragGeneScanRS on {len(samples['nucleotide'])} files")
        start_time = timeit.default_timer()
        outfgs = os.path.join(m_outputfolder, 'fgs')
        jobsFGS = []
        for basename,files in samples['nucleotide'].items():
            jobsFGS += [orf_call_fgs.remote(basename, files[0], outfgs)]
        while jobsFGS:
            ready,jobsFGS = ray.wait(jobsFGS)
            ret = ray.get(ready[0])
            if ret:
                samples['fgs'][ret[0]] = [ret[1]]
        if DEBUG:
            print(f"Time to run FGS: {round(timeit.default_timer() - start_time,2)} seconds")
            print(f"Virtual Memory {mem_use()}GB")


    ## Process Proteins ##
    for sample_type in samples.keys():
        if (not samples[sample_type]) or (sample_type == "nucleotide"):
            continue
        print(f"\nProcessing Proteins ({sample_type})")
        print(f"Running Mercat2 using {m_num_cores} cores")
        # Chunk large files
        if m_chunk_size > 0:
            print("Checking for large protein files")
            start_time = timeit.default_timer()
            dir_chunks = os.path.join(m_outputfolder, f"chunks_{sample_type}")
            jobs = []
            for basename,file in samples[sample_type].items():
                chunk_path = os.path.join(dir_chunks, basename)
                jobs += [chunk_files.remote(basename, file[0], m_chunk_size, chunk_path)]
            while jobs:
                ready,jobs = ray.wait(jobs)
                name,chunks = ray.get(ready[0])
                chunks = [Path(c) for c in chunks]
                samples[sample_type][name] += chunks
        if DEBUG:
            print(f"Time to check for large protein files: {round(timeit.default_timer() - start_time,2)} seconds")

        jobs = []
        tsv_list = dict()
        out_tsv = os.path.join(m_outputfolder, f"tsv_{sample_type}")
        os.makedirs(out_tsv, exist_ok=True)
        start_time = timeit.default_timer()
        for basename,files in samples[sample_type].items():
            out_counts = os.path.join(out_tsv, f"{basename}_counts.tsv")
            files = files[1:] if len(files)>1 else [files[0]]
            jobs += [run_mercat2.remote(basename, files, out_counts, m_kmer, m_min_count, m_num_cores)]
        while(jobs):
            ready,jobs = ray.wait(jobs)
            if ready[0]:
                basename,kmers = ray.get(ready[0])
                if kmers:
                    tsv_list[basename] = kmers
        print(f"Time to count {m_kmer}-mers: {round(timeit.default_timer() - start_time,2)} seconds")
        if DEBUG:
            print(f"Virtual Memory {mem_use()}GB")
        start_time = timeit.default_timer()
        if len(tsv_list):
            figPlots.update(createFigures(tsv_list, sample_type, m_outputfolder, m_lowmem, m_class_file, m_pca))
            tsvfile = Path(m_outputfolder, f'combined_{sample_type}_T.tsv')
            mercat2_diversity.compute_beta_diversity(sample_type, tsvfile, Path(m_outputfolder, "report", "beta_diversity"))
            if DEBUG:
                print(f"Time to calculate Beta Diversity: {round(timeit.default_timer() - start_time,2)} seconds")
        for basename,filename in tsv_list.items():
            os.makedirs(os.path.join(report_dir, 'diversity'), exist_ok=True)
            outfile = os.path.join(report_dir, 'diversity', f'{sample_type}-{basename}.tsv')
            jobsDiversity += [diversity.remote(basename, filename, outfile, sample_type)]


    # Plot Data
    mercat2_report.write_html(os.path.join(report_dir, "report.html"), figPlots, tsv_stats)
    for sample_type in ['protein', 'fgs', 'prod']:
        if len(samples[sample_type]):
            tsv_out = os.path.join(report_dir, f'metrics-{sample_type}.tsv')
            htm_out = os.path.join(report_dir, f'metrics-{sample_type}.html')
            figPlots = mercat2_figures.plot_sample_metrics(samples[sample_type], tsv_out)
            mercat2_report.write_html(htm_out, figPlots, dict())

    # Wait for any remaining QC Jobs
    if jobsQC:
        print("Waiting for any remaining QC jobs")
    while jobsQC:
        ready,jobsQC = ray.wait(jobsQC)

    print("Gathering Diversity Metrics")
    mergedDiversity = dict()
    while jobsDiversity:
        ready,jobsDiversity = ray.wait(jobsDiversity)
        key,outfile = ray.get(ready[0])
        if key not in mergedDiversity:
            mergedDiversity[key] = []
        mergedDiversity[key].append(outfile)

    for key,filelist in mergedDiversity.items():
        if len(filelist) < 2:
            continue
        tomerge = dict()
        for filename in filelist:
            name = re.search(r'\w+-(\w+).tsv', Path(filename).name)
            if name:
                name = name.group(1)
                tomerge[name] = filename
        outfile = Path(report_dir, f'diversity-{key}.tsv')
        mercat2_report.merge_tsv(tomerge, outfile)

    print("\nFinished MerCat2 Pipeline")

    return


# If main app
if __name__ == "__main__":
    mercat_main()



# -*- coding: utf-8 -*-
"""mercat2_fasta.py: Module with tools for fasta files
"""

import os
import sys
from pathlib import Path
import pkg_resources as pkg
import gzip
import shutil
import subprocess
import tarfile
import re
import textwrap

PATH_FGS = pkg.resource_filename("mercat2_lib", "FGS")


## Split Sequence by N
def split_sequenceN(header:str, sequence:str):
    '''Splits a nucleotide sequence string at occurrences of N.
    Helper method for removeN.
    Assumes the first word before the first space is the name of the sequence.

    Parameters:
        header (str): The header of the sequence, without the leading '>'.
        sequence (str): The nucleotide sequence.

    Returns:
        tuple: A tuple with a list of sequences (including modified headers)
               and a corresponding list of contiguous N repeats in each sequence.
    '''

    N_lengths = []
    regex = re.compile(r"(N+)")
    for match in regex.finditer(sequence):
        N_lengths.append(len(match.group(1)))
    sequences = regex.sub('\n', sequence).split('\n')
    header = header.split()
    basename = header[0]
    info = ' '.join(header[1:])
    seqs = []
    for i, seq in enumerate(sequences, 1):
        header = f">{basename}_{i} {info}"
        seqs.append(header)
        seqs += textwrap.wrap(seq, 80)
    
    return (seqs, N_lengths)


## Remove N's
def removeN(fasta:Path, outpath:Path, toupper:bool):
    '''Splits sequences in a scaffold fasta file at N repeats.

    Parameters:
        fasta (Path): A path to a fasta file.
        outpath (str): A folder path of where to save the modified file.

    Returns:
        tuple: A tuple with the path to the cleaned file and a dictionary containing basic stats.
    '''

    os.makedirs(outpath, exist_ok=True)    

    GZIP = Path(fasta).suffix == '.gz'

    basename = Path(fasta).stem.split('.')[0]
    ext = ''.join(Path(fasta).suffixes)
    outFasta = Path(outpath, f"{basename}_clean.fna.gz")

    NStats = dict()
    gc_count = 0
    total_length = 0
    reader = gzip.open(fasta, 'rt') if GZIP else open(fasta, 'r')
    with gzip.open(outFasta, 'wt') as writer:
        line = reader.readline()
        while line:
            line = line.strip()
            if line.startswith('>'):
                name = line[1:]
                sequence = ""
                seq_line = list()
                line = reader.readline()
                while line:
                    line = line.strip()
                    if line.startswith('>'):
                        break
                    sequence += line
                    seq_line += [line]
                    line = reader.readline()
                if 'N' in sequence:
                    sequences, stats = split_sequenceN(name, sequence)
                    #print('\n'.join(sequences), file=writer)
                    for seq in sequences:
                        if seq.startswith('>'):
                            print(seq, file=writer)
                        else:
                            if toupper:
                                print(seq.upper(), file=writer)
                            else:
                                print(seq, file=writer)
                        gc_count += seq.count('G') + seq.count('C')
                        total_length += len(seq)
                else:
                    print('>', name, sep='', file=writer)
                    for seq in seq_line:
                        if toupper:
                            print(seq.upper(), file=writer)
                        else:
                            print(seq, file=writer)
                    gc_count += sequence.count('G') + sequence.count('C')
                    total_length += len(sequence)
                continue #already got next line, next item in loop
            line = reader.readline()
    reader.close()
    NStats['GC Content'] = 100.0 * gc_count / total_length

    return (outFasta.absolute(), NStats)


## Check Command
#
def check_command(cmd:str):
    '''Checks if the command is available in PATH and executable'''

    if shutil.which(cmd) is None:
        print(f"Mercat Error: {cmd} not found, please setup {cmd} using: 'conda install {cmd}'")
        return False
    else:
        return True


## Process Fastq
def qc(fq_file:str, outpath:str, f_name:str):
    '''Processes fastq files.
    Uses fastqc to get the quality of the reads.

    Parameters:
        fq_file (str): The path to a fastq file.
        outpath (str): The path to save the fastqc report.
        f_name (str): The name of the sample.

    Returns:
        None
    '''

    os.makedirs(outpath, exist_ok=True)
    if check_command('fastqc'):
        subprocess.run(['fastqc', fq_file, '-o', outpath], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
    return

def trim(fq_file:str, outpath:str, f_name:str):
    '''Processes fastq files.
    Uses fastp to trim the reads.

    Parameters:
        fq_file (str): The path to a fastq file.
        outpath (str): The path to save the trimmed fastq file.
        f_name (str): The name of the sample.

    Returns:
        str: The path to the trimmed fasta file.
    '''

    os.makedirs(outpath, exist_ok=True)
    trim_fq = os.path.join(outpath, f_name+"_trim.fastq")
    if check_command('fastp'):
        subprocess.run(['fastp', '-i', fq_file, '-o', trim_fq], stdout=open(f'{outpath}/{f_name}-trim.stdout', 'w'), stderr=open(f'{outpath}/{f_name}-trim.stderr', 'w'))
    else:
        print("WARNING: Continuing without trim")
        return fq_file
    return trim_fq

def fq2fa(fq_file:str, outpath:str, f_name:str):
    '''Processes fastq files.
    Uses Linux sed command to convert the fastq file to fasta format.

    Parameters:
        fq_file (str): The path to a fastq file.
        outpath (str): The path to save the  fasta file.
        f_name (str): The name of the sample.

    Returns:
        str: The path to the converted fasta file.
    '''

    os.makedirs(outpath, exist_ok=True)
    fna_file = os.path.join(outpath, f_name+".fna.gz")

    # convert fastq to fasta
    cat = 'zcat' if fq_file.endswith('.gz') else 'cat'
    pcat = subprocess.Popen([cat, fq_file], stdout=subprocess.PIPE)
    proc = subprocess.Popen(['sed', '-n', '1~4s/^@/>/p;2~4p'], stdin=pcat.stdout, stdout=subprocess.PIPE, text=True)
    with gzip.open(fna_file, 'wt') as writer:
        for line in proc.stdout:
            writer.write(line)
    return os.path.abspath(fna_file)


## ORF Call Prodigal
def orf_call(basename:str, fna_in:str, outpath:str):
    '''Finds the ORFs in the nucleotides and converts to an amino acid file.
    Uses prodigal for ORF calling.
    Produces a protein ffa file.
    Produces a gff file.

    Parameters:
        basename (str): The base name of the file for output files.
        fna_in (str): The path to a nucleotide fasta file.
        outpath (str): The path to save the output files.

    Returns:
        str: The path to the protein faa file.
    '''

    if not check_command('prodigal'):
        exit()

    outpath = os.path.abspath(outpath)
    faa_tmp = os.path.join(outpath, basename+"_pro.faa")
    faa_out = faa_tmp # os.path.join(outpath, basename+"_pro.faa.gz")
    prod_cmd = ['prodigal',
                '-a', faa_out,
                '-p', 'meta']
    os.makedirs(outpath, exist_ok=True)
    pcat = subprocess.Popen(['zcat', fna_in], text=True, stdout=subprocess.PIPE)
    with open(f'{outpath}/{basename}.gbk', 'w') as stdout, open(f'{outpath}/{basename}.stderr', 'w') as stderr:
        subprocess.run(prod_cmd, stdout=stdout, stderr=stderr, stdin=pcat.stdout)

    return (basename, faa_out)


## ORF Call FragGeneScanRS
def orf_call_fgs(basename:str, fna_in:str, outpath:str):
    '''Finds the ORFs in the nucleotides and converts to an amino acid file.
    Uses FragGeneScanRS for ORF calling.
    Produces a protein ffa file.
    Produces a gff file.

    Parameters:
        basename (str): The base name of the file for output files.
        fna_in (str): The path to a nucleotide fasta file.
        outpath (str): The path to save the output files.

    Returns:
        tuple: A tuple with the name, and path to the protein faa file.
    '''

    exe_fgs = os.path.join(PATH_FGS, 'FragGeneScanRs')

    if not os.path.exists(exe_fgs):
        system = sys.platform

        if system == "Windows":
            print("Windows is not supported with FragGeneScanRs")
            return None

        with tarfile.open(os.path.join(PATH_FGS, f'FragGeneScanRS-{system}.tar.gz'), 'r') as fgs:
            fgs.extractall(PATH_FGS)

    outpath = os.path.abspath(outpath)
    os.makedirs(outpath, exist_ok=True)
    faa_out = os.path.join(outpath, f'{basename}.faa.gz')

    command = ['FragGeneScanRs',
                '--complete',
                '-t', 'complete',
                ]

    pcat = subprocess.Popen(['zcat', fna_in], text=True, stdout=subprocess.PIPE)
    pout = subprocess.Popen(command, text=True, stdin=pcat.stdout, stdout=subprocess.PIPE)
    with gzip.open(faa_out, 'wt') as writer:
        for line in pout.stdout:
            writer.write(line)

    return (basename, faa_out)
